<?php

namespace Drupal\config_sync;

use Drupal\config_provider\InMemoryStorage;
use Drupal\config_snapshot\ConfigSnapshotStorageTrait;
use Drupal\config_sync\Plugin\SyncConfigCollectorInterface;
use Drupal\Core\Config\ConfigManagerInterface;
use Drupal\Core\Config\StorageComparer;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Extension\Extension;

/**
 * The ConfigSyncSnapshotter provides helper functions for taking snapshots of
 * extension-provided configuration.
 */
class ConfigSyncSnapshotter implements ConfigSyncSnapshotterInterface {

  use ConfigSnapshotStorageTrait;

  /**
   * The app root for the current operation.
   *
   * @var string
   */
  protected $root;

  /**
   * The configuration collector.
   *
   * @var \Drupal\config_sync\Plugin\SyncConfigCollectorInterface
   */
  protected $configCollector;

  /**
   * The provider configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $providerStorage;

  /**
   * The configuration manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected $configManager;

  /**
   * The active configuration store.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $activeStorage;

  /**
   * Constructs a ConfigSyncSnapshotter object.
   *
   * @param string $root
   *   The app root.
   * @param \Drupal\config_provider\Plugin\SyncConfigCollectorInterface $config_collector
   *   The config collector.
   * @param \Drupal\Core\Config\StorageInterface $provider_storage
   *   The provider configuration storage.
   * @param \Drupal\Core\Config\ConfigManagerInterface $config_manager
   *   The configuration manager.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active configuration store where the list of enabled modules and
   *   themes is stored.
   */
  public function __construct($root, SyncConfigCollectorInterface $config_collector, StorageInterface $provider_storage, ConfigManagerInterface $config_manager, StorageInterface $active_storage) {
    $this->root = $root;
    $this->configCollector = $config_collector;
    $this->providerStorage = $provider_storage;
    $this->configManager = $config_manager;
    $this->activeStorage = $active_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function refreshExtensionSnapshot($type, array $names, $mode) {
    foreach ($names as $name) {
      $extensions = [];
      $pathname = $this->drupalGetFilename($type, $name);
      $extensions[$name] = new Extension($this->root, $type, $pathname);
      $snapshot_storage = $this->getConfigSnapshotStorage(ConfigSyncSnapshotterInterface::CONFIG_SNAPSHOT_SET, $type, $name);
      switch ($mode) {
        case ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_INSTALL:
          // Take the snapshot in two stages.
          // First, add the unaltered configuration.
          $this->configCollector->addConfigForSnapshotting($extensions);
          $this->saveSnapshotChanges($snapshot_storage);
          // Now alter the previously added configuration.
          $this->configCollector->alterConfigForSnapshotting($extensions);
          $this->saveSnapshotChanges($snapshot_storage);
          break;
        case ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_IMPORT:
          // No special handling for alter on update since the item will already have
          // been altered on install.
          $this->configCollector->addInstallableConfig($extensions);
          $this->saveSnapshotChanges($snapshot_storage);
          break;
      }
    }
  }

  /**
   * Saves changes from the provider storage to a snapshot storage.
   *
   * @param \Drupal\Core\Config\StorageInterface $snapshot_storage
   *   The storage to synchronize configuration to.
   */
  protected function saveSnapshotChanges(StorageInterface $snapshot_storage) {
    // Set up a storage comparer.
    $storage_comparer = new StorageComparer(
      $this->providerStorage,
      $snapshot_storage,
      $this->configManager
    );
    $storage_comparer->createChangelist();

    $changelist = $storage_comparer->getChangelist();

    // Only add new or changed items.
    foreach (['create', 'update'] as $changelist_type) {
      if (isset($changelist[$changelist_type])) {
        foreach ($changelist[$changelist_type] as $item_name) {
          if ($provided_data = $this->providerStorage->read($item_name)) {
            // Snapshot the configuration item as provided by the extension.
            $snapshot_storage->write($item_name, $provided_data);
          }
        }
      }
    }

    // Snapshot collections as well.
    foreach ($this->providerStorage->getAllCollectionNames() as $collection) {
      $changelist = $storage_comparer->getChangelist(NULL, $collection);
      $this->providerStorage = $this->providerStorage->createCollection($collection);
      $collection_snapshot_storage = $snapshot_storage->createCollection($collection);

      // Only add new or changed items.
      foreach (['create', 'update'] as $changelist_type) {
        if (isset($changelist[$changelist_type])) {
          foreach ($changelist[$changelist_type] as $item_name) {
            if ($provided_data = $this->providerStorage->read($item_name)) {
              // Snapshot the configuration item as provided by the extension.
              $collection_snapshot_storage->write($item_name, $provided_data);
            }
          }
        }
      }

    }

  }

  /**
   * {@inheritdoc}
   */
  public function createSnapshot() {
    $extensions = $this->activeStorage->read('core.extension');
    $types = [
      'module',
      'theme',
    ];
    foreach ($types as $type) {
      if (isset($extensions[$type])) {
        $this->refreshExtensionSnapshot($type, array_keys($extensions[$type]), ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_INSTALL);
      }
    }
  }

  /**
   * Copies the content of one storage to another.
   *
   * @param \Drupal\Core\Config\StorageInterface $source_storage
   *   The storage to synchronize configuration from.
   * @param \Drupal\Core\Config\StorageInterface $snapshot_storage
   *   The storage to synchronize configuration to.
   */
  protected function copyStorage($source_storage, $snapshot_storage) {
    foreach ($source_storage->listAll() as $name) {
      $snapshot_storage->write($name, $source_storage->read($name));
    }
    // Copy collections as well.
    foreach ($source_storage->getAllCollectionNames() as $collection) {
      $source_collection = $source_storage->createCollection($collection);
      $snapshot_collection = $snapshot_storage->createCollection($collection);
      foreach ($source_collection->listAll() as $name) {
        $snapshot_collection->write($name, $source_collection->read($name));
      }
    }
  }

  /**
   * Wrapper for drupal_get_filename().
   *
   * @param $type
   *   The type of the item; one of 'core', 'profile', 'module', 'theme', or
   *   'theme_engine'.
   * @param $name
   *   The name of the item for which the path is requested. Ignored for
   *   $type 'core'.
   *
   * @return string
   *   The path to the requested item or an empty string if the item is not
   *   found.
   */
  protected function drupalGetFilename($type, $name) {
    return drupal_get_filename($type, $name);
  }

}
